/*
 *  Implementation of file statements and functions.
 *  TODO: error checking
 */


#include <stdio.h>
#include <string.h>  // strdup

#include "Object.h"
#include "arg.h"
#include "expr.h"
#include "lex.h"  // nextToken()

void expect(char c);  // in basic.cpp


// ---- File module ----

FILE *channel_to_fp(int channel)
{
  // This is dangerous. A wrong channel number (which doesn't come from
  // OPENXX) would cause the interpreter to crash.
  return (FILE *) channel;
}


/*
 *  Reads an integer factor and converts it to the corresponding file handle.
 */
FILE *getFP()
{
  // expect('#');
  int channel = int_factor();
  return (FILE *) channel;  // dangerous!
}


void exec_bget(Object *res)
{
  FILE *c = getFP();
  res->setInt(fgetc(c));
}


void exec_bput()
{
  Object val;

  FILE *c = getFP();
  expect(',');
  expression(&val);
  // check errors!
  if(val.type == STRING) {
    fputs(val.stringValue, c);
    if(ttype == ';')
      nextToken();
    else
      fputc('\n', c);
  }
  else
    fputc(val.toInt(), c);
}


void exec_close()
{
  FILE *c = getFP();
  fclose(c);
}


void exec_eof(Object *val)
{
  FILE *c = getFP();
  val->setInt(feof(c) ? -1 : 0);
}


void exec_ext(Object *val)
{
  // Maybe this could be implement more elegantly with system dependent
  // calls (like UNIX stat()). But then we would need
  //   - the file name / file descriptor for each channel
  //   - implementations for each system

  FILE *c = getFP();
  long pos = ftell(c);      // store current position
  fseek(c, 0L, SEEK_END);   // goto end of file
  long ext = ftell(c);      // position = file length
  fseek(c, pos, SEEK_SET);  // restore old position

  val->setInt((int) ext);
}


void exec_getString(Object *val)
{
  char buf[256];

  FILE *c = getFP();
  // fgets reads until newline, not CR or NULL!
  fgets(buf, 256, c);
  val->setString(strdup(buf));  // strdup here or in setString() ?
}


void exec_openin(Object *res)
{
  char *filename = string_factor();
  FILE *c = fopen(filename, "r");
  // returns 0 if failed
  res->setInt((int) c);
}


void exec_openout(Object *res)
{
  char *filename = string_factor();
  FILE *c = fopen(filename, "w+");
  // returns 0 if failed
  res->setInt((int) c);
}


void exec_openup(Object *res)
{
  char *filename = string_factor();
  FILE *c = fopen(filename, "r+");
  // returns 0 if failed
  // what should be returned by OPENUP if file doesn't exist?
  res->setInt((int) c);
}


// function context

void exec_ptr(Object *res)
{
  FILE *c = getFP();
  long pos = ftell(c);
  res->setInt((int) pos);
}


// statement context: PTR#channel = expression

void exec_ptr()
{
  FILE *c = getFP();
  expect('=');
  int pos = int_expression();
  fseek(c, pos, SEEK_SET);
}

