/*
 * Operator functions
 */

#include <math.h>    // pow()
#include <string.h>  // strcat()

#include "token.h"  // we only need the operator tokens
#include "error.h"
#include "Object.h"

#define boolean int



void unaryOperator(int op, Object *a)
{
  if(a->isString())
    error(6, "Type mismatch: number needed");

  // '+' can be ignored (identity)
  if(op == '-') {
    if(a->isInteger())
      a->intValue = - a->intValue;
    else
      a->floatValue = - a->floatValue;
  }
  else if(op == TT_NOT)
    a->setInt(~ a->toInt());
}




void numericOperator(int op, Object *a, Object *b)
{
  float fres;  // wieso hier??
  int ires;
  boolean bothInt = a->isInteger() && b->isInteger();

  switch(op) {
  case '^':
    // consider integer calculation...
    fres = pow(a->toFloat(), b->toFloat());
    if(bothInt)
      a->intValue = (int) fres;
    else
      a->setFloat(fres);
    break;
  case '*':
    if(bothInt)
      a->intValue *= b->intValue;
    else
      a->setFloat(a->toFloat() * b->toFloat());
    break;
  case '/':
    if(b->toFloat() == 0)
      error(1, "Division by zero");
    // a division always returns a float result
    a->setFloat(a->toFloat() / b->toFloat());
    break;
  case TT_DIV:
    if(b->toFloat() == 0)
      error(1, "Division by zero");
    a->setInt(a->toInt() / b->toInt());
    break;
  case TT_MOD:
    if(b->toFloat() == 0)
      error(1, "Division by zero");
    a->setInt(a->toInt() % b->toInt());
    break;
  case '+':
    if(bothInt)
      a->intValue += b->intValue;
    else
      a->setFloat(a->toFloat() + b->toFloat());
    break;
  case '-':
    if(bothInt)
      a->intValue -= b->intValue;
    else
      a->setFloat(a->toFloat() - b->toFloat());
    break;
  case '=':
    if(bothInt)
      ires = a->intValue == b->intValue;
    else
      ires = a->toFloat() == b->toFloat();
    a->setInt(ires ? -1 : 0);
    break;
  case TT_NE:
    if(bothInt)
      ires = a->intValue != b->intValue;
    else
      ires = a->toFloat() != b->toFloat();
    a->setInt(ires ? -1 : 0);
    break;
  case TT_LE:
    if(bothInt)
      ires = a->intValue <= b->intValue;
    else
      ires = a->toFloat() <= b->toFloat();
    a->setInt(ires ? -1 : 0);
    break;
  case TT_GE:
    if(bothInt)
      ires = a->intValue >= b->intValue;
    else
      ires = a->toFloat() >= b->toFloat();
    a->setInt(ires ? -1 : 0);
    break;
  case '<':
    if(bothInt)
      ires = a->intValue < b->intValue;
    else
      ires = a->toFloat() < b->toFloat();
    a->setInt(ires ? -1 : 0);
    break;
  case '>':
    if(bothInt)
      ires = a->intValue > b->intValue;
    else
      ires = a->toFloat() > b->toFloat();
    a->setInt(ires ? -1 : 0);
    break;
  default:
    error(0, "Not implemented: %c", op);
  }
}


/*
 * With strings, only concatenation (+) and comparisons are allowed.
 */
void stringOperator(int op, Object *a, Object *b) {
  if(op == '+')
    // ---- string concatenation ----
    strcat(a->stringValue, b->stringValue);  // check size!
  else {
    // ---- string comparison ----
    int res;
    int cmp = strcmp(a->stringValue, b->stringValue);
    switch(op) {
    case '=':
      res = cmp == 0;
      break;
    case TT_NE:
      res = cmp != 0;
      break;
    case '<':
      res = cmp < 0;
      break;
    case '>':
      res = cmp > 0;
      break;
    case TT_LE:
      res = cmp <= 0;
      break;
    case TT_GE:
      res = cmp >= 0;
      break;
    }
    a->setInt(res);  // is this -1 or 0?
  }
}


void binaryOperator(int op, Object *a, Object *b)
{
  if(a->isString() && !(b->isString()) ||
     b->isString() && !(a->isString()))
    error(1, "Incompatible types");

  if(a->isString()) {
    if(op == '+' || op == '=' || op == '<' || op == '>' ||
       op == TT_NE || op == TT_LE || op == TT_GE)
      stringOperator(op, a, b);
    else
      // operation not allowed with strings
      error(1, "Type mismatch: number needed");
  }
  else
    // both operands are numeric
    numericOperator(op, a, b);
}
