/*
 * Statement handlers
 */

#include <stdio.h>

#include "Object.h"
#include "SymbolTable.h"
#include "error.h"
#include "bool.h"
#include "lex.h"
#include "token.h"
#include "expr.h"
#include "arg.h"

#include "basic.h"


// for loop stuff

Object *for_var;
float for_limit;
float for_step;
int for_lineno;

int repeat_lineno;


void exec_load()
{
  Object val;

  expression(&val);
  if(val.type != STRING)
    error(6, "Type mismatch: string needed");

  load(val.stringValue);
}




void goto_line(int n)
{
//   lineno = n;
//   lex_start(lines[lineno]);
  jump_to(n, lines[n]);
}


/*
 * Executes statements until a block closing statement is found, i. e.
 * ELSE, ENDCASE, ENDIF, ENDWHILE, NEXT, OTHERWISE, UNTIL, WHEN, END,
 * ENDPROC, =..., RETURN
 */
void interpret_block()
{
  do {
    if(interpret_statement())
      return;
    if(ttype == TT_EOL) {
      // next line (check for end...)
      lineno++;
      if(!lines[lineno]) {
	// last line
	end = true;
	return;
      }
      lex_start(lines[lineno]);
      line = lines[lineno];
      nextToken();
    }
    if(ttype == ':')
      nextToken();
  }
  while(true);
}


void exec_repeat_until()
{
  int repeat_lineno;
  float cond;

  nextToken();  // skip REPEAT, expect EOL
  repeat_lineno = lineno;
  do {
    lineno = repeat_lineno;
    interpret_block();
    if(ttype != TT_UNTIL)
      error(1, "Missing UNTIL"); // bzw. Not in ...
    nextToken();
    cond = float_expression();
  }
  while(cond == 0);

}


void exec_for_next()
{
  Object *for_var;
  float for_limit;
  float for_step;
  int for_lineno;

  nextToken();
  if(ttype != TT_IDENTIFIER)
    error(1, "Identifier expexted after FOR");
  if(isString(sval))
    error(6, "Type mismatch: numeric variable needed");
  //strcpy(for_var, sval);  // save varname
  for_var = symbols.lookup_or_create(sval);
  nextToken();
  expect('=', 1, "Missing = in FOR statement");

  Object val;
  expression(&val);
  assign(for_var, &val);
  expect(TT_TO, 1, "Missing TO");
  for_limit = float_expression();
  if(ttype == TT_STEP) {
    nextToken();
    // distinguish between int and float...
    for_step = float_expression();
  }
  else
    for_step = 1;

  for_lineno = lineno;

  while(true) {
    interpret_block();
    
    if(ttype != TT_NEXT)
      error(1, "Missing NEXT"); // bzw. Not in ...
    nextToken();
    
    // var += step
    if(for_var->isInteger())
      for_var->intValue += (int) for_step;
    else
      for_var->floatValue += for_step;

    if(for_var->toFloat() > for_limit)
      return;  // oder break

    //goto_line(for_lineno);
  }
}


// void exec_repeat()
// {
//   nextToken();
//   repeat_lineno = lineno;
// }


// void exec_until()
// {
//   Object val;

//   nextToken();
//   expression(&val);
//   // check for string...
//   if(val.toFloat() == 0)
//     lineno = repeat_lineno;
// }


void jump_to(int token)
{
  do {
    lineno++;  // next line
    lex_start(lines[lineno]);
    nextToken();
  }
  while(ttype != token);
  nextToken();  // should be TT_EOL, checked at end of interpret()
}




void exec_lvar()
{
  symbols.lvar();  // not nice...
}



// ---- Graphics module ----

static int xpos, ypos;

void exec_move()
{
  bool relative;

  if(ttype == 0 /*TT_BY*/) {
    relative = true;
    nextToken();
  }

  int x = int_expression();
  expect(',');
  int y = int_expression();

  // exec

  if(relative) {
    xpos += x;
    ypos += y;
  }
  else {
    xpos = x;
    ypos = y;
  }
}


// ---- Trigonometric module ----

// cosine function
void eval_cos(Object *res)
{
  float arg = float_factor();
  res->floatValue = cos(arg);
  res->type = FLOAT;
}

